import zmq
import uuid
import time
from hkube_python_wrapper.util.logger import log
from hkube_python_wrapper.communication.zmq.streaming import signals

context = zmq.Context()
POLL_MS = 1000
MAX_POLLS = 5

class ZMQListener(object):
    def __init__(self, remoteAddress, onMessage, encoding, consumerType):
        self._encoding = encoding
        self._onMessage = onMessage
        self._consumerType = self._encoding.encode(consumerType, plainEncode=True)
        self._active = True
        self._working = True
        self._pollTimeoutCount = 0
        self._remoteAddress = remoteAddress
        self._worker = self._worker_socket(remoteAddress)

    def _worker_socket(self, remoteAddress):
        """Helper function that returns a new configured socket
           connected to the Paranoid Pirate queue"""
        identity = str(uuid.uuid4()).encode()
        worker = context.socket(zmq.DEALER)
        worker.setsockopt(zmq.IDENTITY, identity)
        worker.connect(remoteAddress)
        log.info("zmq listener connecting to {addr}", addr=remoteAddress)
        return worker

    def _send(self, signal, result=None):
        arr = [signal, self._consumerType, result or signals.PPP_EMPTY]
        self._worker.send_multipart(arr, copy=False)

    def _handleAMessage(self, frames):
        _, encodedMessageFlowPattern, header, message = frames  # pylint: disable=unbalanced-tuple-unpacking
        messageFlowPattern = self._encoding.decode(value=encodedMessageFlowPattern, plainEncode=True)
        return self._onMessage(messageFlowPattern, header, message)

    def fetch(self):
        try:
            if (self._active is False):
                time.sleep(0.2)
                return

            if (self._pollTimeoutCount == MAX_POLLS):
                log.warning('ZMQListener poll timeout reached')
                self._pollTimeoutCount = 0
                self._worker.close()
                self._worker = self._worker_socket(self._remoteAddress)

            if (self._pollTimeoutCount > 0):
                self._readMessage()
                return

            self._send(signals.PPP_READY)
            self._readMessage()

        except Exception as e:
            log.error('ZMQListener.fetch {e}', e=str(e))
        finally:
            if(self._active is False):
                self._working = False

    def _readMessage(self, timeout=POLL_MS):
        hasMsg = False
        result = self._worker.poll(timeout)
        if (result == zmq.POLLIN):
            self._pollTimeoutCount = 0
            frames = self._worker.recv_multipart()
            signal = frames[0]

            if (signal == signals.PPP_MSG):
                hasMsg = True
                msgResult = self._handleAMessage(frames)
                self._send(signals.PPP_DONE, msgResult)
            else:
                time.sleep(0.005)
        else:
            self._pollTimeoutCount += 1
            log.warning('ZMQListener poll timeout {count}', count=self._pollTimeoutCount)
        return hasMsg

    def close(self, force=True):
        closed = False
        if (self._active is False):
            log.warning('attempting to close inactive ZMQListener')
        else:
            self._active = False
            while self._working and not force:
                time.sleep(0.02)

            if (self._pollTimeoutCount):
                log.warning('trying to read message from socket after close')
                hasMsg = self._readMessage(timeout=POLL_MS * 5)
                if (hasMsg):
                    log.warning('success reading message from socket after close')

            self._worker.close()
            closed = True
        return closed



from hkube_python_wrapper.communication.streaming.StreamingManager import StreamingManager
from hkube_python_wrapper.communication.streaming.MessageListener import MessageListener
from hkube_python_wrapper.communication.streaming.MessageProducer import MessageProducer
import time

parsedFlows = {
    'analyze': [{'source': 'A', 'next': ['B']}, {'source': 'B', 'next': ['C']}, {'source': 'C', 'next': ['D']}], 
    'master': [{'source': 'B', 'next': ['A', 'C']}, {'source': 'C', 'next': ['D']}, {'source': 'D', 'next': ['E']}]
}

def test_streaming_flow():
    messages = []

    def onMessage(msg, origin):
        messages.append(msg)

    def statsInvoked(args):
        print('stats')

    port = 9340
    nodeName = 'B'
    parents = [{'nodeName': 'A', 'address': {'host': '127.0.0.1', 'port': port}, 'type': 'Add'}]
    producer_config = {'port': port, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
    listen_config = {'encoding': 'msgpack'}
    streamingManager = StreamingManager()
    streamingManager.setParsedFlows(parsedFlows, 'analyze')
    streamingManager.setupStreamingProducer(statsInvoked, producer_config, [nodeName], 'A')
    streamingManager.setupStreamingListeners(listen_config, parents, nodeName)
    streamingManager.registerInputListener(onMessage)
    streamingManager.startMessageListening()

    streamingManager.sendMessage({'msg': '1'})
    streamingManager.sendMessage({'msg': '2'})
    streamingManager.sendMessage({'msg': '3'})
    streamingManager.sendMessage({'msg': '4'})

    time.sleep(1)

    assert len(messages) == 4
    assert messages[0] == {'msg': '1'}
    assert messages[1] == {'msg': '2'}
    assert messages[2] == {'msg': '3'}
    assert messages[3] == {'msg': '4'}
    streamingManager.stopStreaming()

def test_streaming_manager():
    resultsAtB = {}
    resultsAtC = {}

    def onMessageAtC(msg, origin):
        resultsAtC['data'] = msg
        resultsAtC['origin'] = origin

    def onMessageAtB(msg, origin):
        resultsAtB['data'] = msg
        resultsAtB['origin'] = origin
        streamingManagerB.sendMessage('msg from B')

    def statsInvoked(args):
        print('stats')

    port1 = 9341
    port2 = 9342
    parents1 = [{'nodeName': 'A', 'address': {'host': '127.0.0.1', 'port': port1}, 'type': 'Add'}]
    parents2 = [{'nodeName': 'B', 'address': {'host': '127.0.0.1', 'port': port2}, 'type': 'Add'}]
    producer_configA = {'port': port1, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
    producer_configB = {'port': port2, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
    listen_config = {'encoding': 'msgpack'}

    streamingManagerA = StreamingManager()
    streamingManagerA.setParsedFlows(parsedFlows, 'analyze')
    streamingManagerA.setupStreamingProducer(statsInvoked, producer_configA, ['B'], 'A')

    streamingManagerB = StreamingManager()
    streamingManagerB.setParsedFlows(parsedFlows, 'analyze')
    streamingManagerB.setupStreamingProducer(statsInvoked, producer_configB, ['C'], 'B')
    streamingManagerB.setupStreamingListeners(listen_config, parents1, 'B')
    streamingManagerB.registerInputListener(onMessageAtB)
    streamingManagerB.startMessageListening()

    streamingManagerC = StreamingManager()
    streamingManagerC.setupStreamingListeners(listen_config, parents2, 'C')
    streamingManagerC.registerInputListener(onMessageAtC)
    streamingManagerC.startMessageListening()

    streamingManagerA.sendMessage('msg from A')
    time.sleep(1)
    assert resultsAtB['data'] == 'msg from A'
    assert resultsAtB['origin'] == 'A'
    assert resultsAtC['data'] == 'msg from B'
    assert resultsAtC['origin'] == 'B'
    streamingManagerA.stopStreaming()
    streamingManagerB.stopStreaming()

def xtest_messaging():

    producer_config = {'port': 9536, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
    listenr_config = {'remoteAddress': 'tcp://localhost:9536', 'encoding': 'msgpack', 'messageOriginNodeName': 'b'}
    asserts = {}
    asserts['responses'] = 0
    messageProducer = MessageProducer(producer_config, ['a'], 'b')

    def onStatistics(statistics):
        asserts['stats'] = statistics
        asserts['responses'] = int(statistics[0]['responses'])

    messageProducer.registerStatisticsListener(onStatistics)
    time.sleep(3)

    def onMessage(envelope, msg, origin):
        if (type(msg) == type(dict())):
            asserts['field1'] = msg['field1']
        time.sleep(0.1)

    messageProducer.start()
    env = [{
        "source": "b",
        "next": [
            "a"
        ]
    }, {
        "source": "a",
        "next": [
            "c"
        ]
    }]
    messageProducer.produce(env, {'field1': 'value1'})
    messageProducer.produce(env, {'field1': 'value1'})
    messageProducer.produce(env, {'field1': 'value1'})
    time.sleep(2)
    assert asserts['stats'][0]['queueSize'] == 3
    assert asserts['stats'][0]['sent'] == 0
    messageListener = MessageListener(listenr_config, receiverNode='a')
    messageListener.registerMessageListener(onMessage)
    for _ in range(1,5):
        if (asserts.get('field1')):
            break
        time.sleep(1)
    time.sleep(4)
    assert asserts['field1'] == 'value1'
    assert asserts['stats'][0]['queueSize'] == 0
    assert asserts['stats'][0]['sent'] == 3
    assert asserts['responses'] == 3

    def getHello(envelope, msg, origin):
        if (msg == b'Hello'):
            asserts['gotHello'] = True
            asserts['envelope'] = envelope

    messageListener.registerMessageListener(getHello)
    messageProducer.produce(env, b'Hello')
    time.sleep(3)
    assert asserts['gotHello']
    assert len(asserts['envelope']) == 1
    assert asserts['envelope'][0]['source'] == 'a'

def xtest_messaging_split():
    producer_config = {'port': 9536, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
    listenr_config_c = {'remoteAddress': 'tcp://localhost:9536', 'encoding': 'msgpack', 'messageOriginNodeName': 'c'}
    asserts = {}
    asserts['responses'] = 0
    messageProducer = MessageProducer(producer_config, ['a','b'], 'c')

    def onStatistics(statistics):
        asserts['stats'] = statistics
        asserts['responses'] = int(statistics[0]['responses'])

    messageProducer.registerStatisticsListener(onStatistics)
    time.sleep(3)

    def onMessage(envelope, msg, origin):
        # pylint: disable=unused-argument
        if (type(msg) == type(dict())):
            asserts['field1'] = msg['field1']
        time.sleep(0.1)

    messageProducer.start()
    env = [{
        "source": "c",
        "next": [
            "a"
        ]
    }]
    env2 = [{
        "source": "c",
        "next": [
            "b"
        ]
    }]
    messageProducer.produce(env, {'field1': 'value1'})
    messageProducer.produce(env2, {'field1': 'value2'})
    messageProducer.produce(env, {'field1': 'value1'})
    messageProducer.produce(env2, {'field1': 'value2'})
    messageProducer.produce(env2, {'field1': 'value2'})
    messageProducer.produce(env, {'field1': 'value1'})
    time.sleep(2)
    assert asserts['stats'][0]['queueSize'] == 3
    assert asserts['stats'][0]['sent'] == 0

    messageListener = MessageListener(listenr_config_c, receiverNode='a')
    messageListener.registerMessageListener(onMessage)
    for _ in range(1,5):
        if ( asserts.get('field1')):
            break
        time.sleep(1)
    time.sleep(2)
    assert asserts['field1'] == 'value1'
    assert asserts['stats'][0]['queueSize'] == 0
    assert asserts['stats'][0]['sent'] == 3
    assert asserts['stats'][1]['queueSize'] == 3
    assert asserts['stats'][1]['sent'] == 0
    assert asserts['responses'] == 3
    messageListener2 = MessageListener(listenr_config_c, receiverNode='b')
    messageListener2.registerMessageListener(onMessage)
    messageListener2.start()
    time.sleep(20)
    assert asserts['stats'][1]['queueSize'] == 0
    assert asserts['stats'][1]['sent'] == 3
    assert len(messageProducer.adapter.messageQueue.queue) == 0



#
# Based on Paranoid Pirate queue
#
#   Author: Daniel Lundin <dln(at)eintr(dot)org>
#
import time
from .Flow import Flow
from .MessageQueue import MessageQueue
from hkube_python_wrapper.util.logger import log
from hkube_python_wrapper.communication.zmq.streaming import signals
import zmq

HEARTBEAT_LIVENESS = 5
CYCLE_LENGTH_MS = 1000

context = zmq.Context()

class ZMQProducer(object):
    def __init__(self, port, maxMemorySize, responseAccumulator, queueTimeAccumulator, consumerTypes, encoding, nodeName):
        self.nodeName = nodeName
        self.encoding = encoding
        self.responseAccumulator = responseAccumulator
        self.queueTimeAccumulator = queueTimeAccumulator
        self.maxMemorySize = maxMemorySize
        self.port = port
        self.consumerTypes = consumerTypes
        self.messageQueue = MessageQueue(consumerTypes, self.nodeName)
        self._backend = context.socket(zmq.ROUTER)  # ROUTER
        self._backend.bind("tcp://*:" + str(port))  # For workers
        log.info("Producer listening on tcp://*:{port}", port=port)
        self._active = True
        self._working = True
        self.watingForResponse = {}

    def produce(self, header, message, messageFlowPattern=[]):
        while (self.messageQueue.sizeSum > self.maxMemorySize):
            self.messageQueue.loseMessage()
        appendTime = time.time()
        self.messageQueue.append(messageFlowPattern, header, message, appendTime)

    def start(self):  # pylint: disable=too-many-branches
        poll_workers = zmq.Poller()
        poll_workers.register(self._backend, zmq.POLLIN)

        while self._active:  # pylint: disable=too-many-nested-blocks
            try:
                socks = dict(poll_workers.poll(CYCLE_LENGTH_MS))

                if socks.get(self._backend) == zmq.POLLIN:

                    frames = self._backend.recv_multipart() or []

                    if(len(frames) != 4):
                        log.warning("got {len} frames {frames}", len=len(frames), frames=frames)
                        continue

                    address, signal, consumer, result = frames # pylint: disable=unbalanced-tuple-unpacking
                    consumerType = self.encoding.decode(value=consumer, plainEncode=True)

                    if (not consumerType in self.consumerTypes):
                        log.warning("Producer got message from unknown consumer: {consumerType}, dropping the message", consumerType=consumerType)
                        continue

                    if(signal == signals.PPP_DONE):
                        sentTime = self.watingForResponse.get(address)
                        if (sentTime):
                            now = time.time()
                            del self.watingForResponse[address]
                            self.responseAccumulator(result, consumerType, round((now - sentTime) * 1000, 4))
                        else:
                            log.error('missing from watingForResponse:' + str(signal))

                    elif (signal == signals.PPP_READY):
                        message = self.messageQueue.pop(consumerType)
                        if (message):
                            messageFlowPattern, header, payload, appendTime = message
                            flow = Flow(messageFlowPattern)
                            flowMsg = self.encoding.encode(flow.getRestOfFlow(self.nodeName), plainEncode=True)
                            frames = [address, signals.PPP_MSG, flowMsg, header, payload]
                            self.watingForResponse[address] = time.time()
                            queueTime = round((time.time() - appendTime) * 1000, 4)
                            self.queueTimeAccumulator(consumerType, queueTime)
                            self._send(frames)
                        else:
                            self._send([address, signals.PPP_NO_MSG])

            except Exception as e:
                log.error('Error in ZMQProducer {e}', e=str(e))

        self._working = False

    def _send(self, frames):
        self._backend.send_multipart(frames, copy=False)

    def queueSize(self, consumerSize):
        return self.messageQueue.size(consumerSize)

    def sent(self, consumerType):
        return self.messageQueue.sent[consumerType]

    def resetQueue(self):
        self.messageQueue.resetAll()

    def close(self, force=True):
        log.info('queue size before close = {len}', len=len(self.messageQueue.queue))
        while self.messageQueue.queue and not force:
            log.info('queue size during close = {len}', len=len(self.messageQueue.queue))
            time.sleep(1)
        self._active = False
        while self._working:
            time.sleep(1)
        self._backend.close()
        log.info('queue size after close = {len}', len=len(self.messageQueue.queue))



import time
from hkube_python_wrapper.util.DaemonThread import DaemonThread

class StreamingListener(DaemonThread):

    def __init__(self, messageListeners):
        self._listeningToMessages = True
        self._working = True
        self._messageListeners = messageListeners
        DaemonThread.__init__(self, "StreamingListener")

    def run(self):
        while (self._listeningToMessages):
            messageListeners = self._messageListeners()
            if (not messageListeners):
                time.sleep(1)  # free some cpu
                continue
            for listener in messageListeners:
                listener.fetch()
        self._working = False

    def stop(self, force=True):
        messageListeners = self._messageListeners()
        for listener in messageListeners:
            listener.close(force)
        self._listeningToMessages = False
        while self._working:
            time.sleep(0.2)



import threading
from .MessageListener import MessageListener
from .MessageProducer import MessageProducer
from .StreamingListener import StreamingListener
from hkube_python_wrapper.util.logger import log


class StreamingManager():

    def __init__(self):
        self.threadLocalStorage = threading.local()
        self.threadLocalStorage.sendMessageId = None
        self.messageProducer = None
        self._messageListeners = dict()
        self._inputListener = []
        self.listeningToMessages = False
        self._isStarted = False
        self.parsedFlows = {}
        self.defaultFlow = None
        self._streamingListener = None


    def setParsedFlows(self, flows, defaultFlow):
        self.parsedFlows = flows
        self.defaultFlow = defaultFlow

    def setupStreamingProducer(self, onStatistics, producerConfig, nextNodes, nodeName):
        self.messageProducer = MessageProducer(producerConfig, nextNodes, nodeName)
        self.messageProducer.registerStatisticsListener(onStatistics)
        if (nextNodes):
            self.messageProducer.start()

    def resetQueue(self):
        self.messageProducer.resetQueue()

    def setupStreamingListeners(self, listenerConfig, parents, nodeName):
        log.debug("parents {parents}", parents=str(parents))
        for parent in parents:
            parentName = parent['nodeName']
            remoteAddress = parent['address']
            remoteAddressUrl = 'tcp://{host}:{port}'.format(host=remoteAddress['host'], port=remoteAddress['port'])

            if (parent['type'] == 'Add'):
                options = {}
                options.update(listenerConfig)
                options['remoteAddress'] = remoteAddressUrl
                options['messageOriginNodeName'] = parentName
                listener = MessageListener(options, nodeName)
                listener.registerMessageListener(self.onMessage)
                self._messageListeners[remoteAddressUrl] = listener

            if (parent['type'] == 'Del'):
                listener = self._messageListeners.get(remoteAddressUrl)
                if (listener):
                    closed = listener.close(force=False)
                    if (closed):
                        self._messageListeners.pop(remoteAddressUrl, None)

    def registerInputListener(self, onMessage):
        self._inputListener.append(onMessage)

    def onMessage(self, messageFlowPattern, msg, origin, sendMessageId=None):
        self.threadLocalStorage.sendMessageId = sendMessageId
        self.threadLocalStorage.messageFlowPattern = messageFlowPattern
        if (not self._inputListener):
            log.error('no input listeners on _onMessage method')
            return
        for listener in self._inputListener:
            try:
                listener(msg, origin)
            except Exception as e:
                log.error("hkube_api message listener threw exception: {e}", e=str(e))
        self.threadLocalStorage.messageFlowPattern = []

    def _getMessageListeners(self):
        return list(self._messageListeners.values())

    def startMessageListening(self):
        self.listeningToMessages = True
        if (self._isStarted is False):
            self._isStarted = True
            self._streamingListener = StreamingListener(self._getMessageListeners)
            self._streamingListener.start()

    def sendMessage(self, msg, flowName=None):
        if (self.messageProducer is None):
            raise Exception('Trying to send a message from a none stream pipeline or after close had been applied on algorithm')
        if (self.messageProducer.nodeNames):
            parsedFlow = None
            if (flowName is None):
                if hasattr(self.threadLocalStorage, 'messageFlowPattern') and self.threadLocalStorage.messageFlowPattern:
                    parsedFlow = self.threadLocalStorage.messageFlowPattern
                else:
                    if (self.defaultFlow is None):
                        raise Exception("Streaming default flow is None")
                    flowName = self.defaultFlow
            if not (parsedFlow):
                parsedFlow = self.parsedFlows.get(flowName)
            if (parsedFlow is None):
                raise Exception("No such flow " + flowName)
            self.messageProducer.produce(parsedFlow, msg)
        else:
            log.error("messageProducer has no consumers")

    def stopStreaming(self, force=True):
        if (self.listeningToMessages):
            self._isStarted = False
            self.listeningToMessages = False
            if (self._streamingListener):
                self._streamingListener.stop(force)
            self._messageListeners = dict()
            self._inputListener = []

        if (self.messageProducer is not None):
            self.messageProducer.close(force)
            self.messageProducer = None

    def clearMessageListeners(self):
        self._messageListeners = dict()



import sys
from events import Events


def stdout_redirector():
    # capture all outputs to a log file while still printing it
    class Logger:
        events = Events()
        _buffer = []
        _sending = False
        _lineBuffer = ''

        def __init__(self, bufferSize=15):
            self.terminal = sys.stdout
            self._bufferSize = bufferSize
            self._stdout = sys.stdout
            self._stderr = sys.stderr
            sys.stdout = self
            sys.stderr = self

        def _send(self):
            self._sending = True
            try:
                tmp = self._buffer
                self._buffer = []
                self.events.on_data(tmp)
            finally:
                self._sending = False

        def write(self, message):
            self.terminal.write(message)
            self._lineBuffer += message
            if ('\n' in self._lineBuffer or '\r' in self._lineBuffer):
                lines = self._lineBuffer.splitlines()
                self._lineBuffer = ''
                for line in lines:
                    self._buffer.append(line)


            if (self.events and not self._sending and len(self._buffer) > self._bufferSize):
                self._send()

        def flush(self):
            if (self.events and not self._sending):
                self._send()

        def __getattr__(self, attr):
            return getattr(self.terminal, attr)

        def cleanup(self):
            sys.stdout = self._stdout
            sys.stderr = self._stderr

    return Logger()
